import logging
from io import StringIO

from django.core.management import call_command
from django.utils import timezone

import pytest
import requests
from freezegun import freeze_time

from log_outgoing_requests.config_reset import schedule_config_reset
from log_outgoing_requests.models import OutgoingRequestsLog, OutgoingRequestsLogConfig

try:
    import celery
except ImportError:
    celery = None

has_celery = celery is not None


@pytest.mark.django_db
def test_cleanup_request_logs_command(settings, caplog):
    settings.LOG_OUTGOING_REQUESTS_MAX_AGE = 1  # delete if > 1 day old

    with freeze_time("2023-10-02T12:00:00Z") as frozen_time:
        OutgoingRequestsLog.objects.create(timestamp=timezone.now())
        frozen_time.move_to("2023-10-04T12:00:00Z")
        recent_log = OutgoingRequestsLog.objects.create(timestamp=timezone.now())

        stdout = StringIO()
        with caplog.at_level(logging.INFO):
            call_command(
                "prune_outgoing_request_logs", stdout=stdout, stderr=StringIO()
            )

    output = stdout.getvalue()
    assert output == "Deleted 1 outgoing request log(s)\n"

    assert OutgoingRequestsLog.objects.get() == recent_log


@pytest.mark.skipif(not has_celery, reason="Celery is optional dependency")
@pytest.mark.django_db
def test_cleanup_request_logs_celery_task(requests_mock, settings, caplog):
    from log_outgoing_requests.tasks import prune_logs

    settings.LOG_OUTGOING_REQUESTS_MAX_AGE = 1  # delete if > 1 old old

    with freeze_time("2023-10-02T12:00:00Z") as frozen_time:
        OutgoingRequestsLog.objects.create(timestamp=timezone.now())
        frozen_time.move_to("2023-10-04T12:00:00Z")
        recent_log = OutgoingRequestsLog.objects.create(timestamp=timezone.now())

        with caplog.at_level(logging.INFO):
            prune_logs()

    assert len(caplog.records) == 1
    assert "Deleted 1 outgoing request log(s)" in caplog.text

    assert OutgoingRequestsLog.objects.get() == recent_log


@pytest.mark.skipif(not has_celery, reason="Celery is optional dependency")
@pytest.mark.django_db
def test_reset_config(requests_mock, settings):
    from log_outgoing_requests.tasks import reset_config

    settings.LOG_OUTGOING_REQUESTS_DB_SAVE = False
    config = OutgoingRequestsLogConfig.get_solo()
    assert isinstance(config, OutgoingRequestsLogConfig)
    config.save_to_db = "yes"
    config.save()
    qs = OutgoingRequestsLog.objects.all()
    requests_mock.get("https://example.com", status_code=200)

    requests.get("https://example.com")
    assert qs.count() == 1

    reset_config()

    requests.get("https://example.com")
    assert qs.count() == 1


@pytest.mark.django_db
def test_saving_config_schedules_config_reset(mocker):
    config = OutgoingRequestsLogConfig.get_solo()
    assert isinstance(config, OutgoingRequestsLogConfig)
    config.save_to_db = "yes"
    mock_schedule = mocker.patch("log_outgoing_requests.models.schedule_config_reset")
    config.save()
    mock_schedule.assert_called_once()


@pytest.mark.skipif(not has_celery, reason="Celery is optional dependency")
def test_schedule_config_schedules_celery_task(settings, mocker):
    mock_task = mocker.patch(
        "log_outgoing_requests.config_reset.reset_config.apply_async"
    )
    settings.LOG_OUTGOING_REQUESTS_RESET_DB_SAVE_AFTER = 1
    schedule_config_reset()
    mock_task.assert_called_once_with(countdown=60)



import logging
import textwrap
from typing import Union, cast

from requests import RequestException
from requests.models import PreparedRequest, Response

from .compat import format_exception
from .typing import RequestLogRecord, is_request_log_record


def format_headers(headers) -> str:
    return "\n".join(f"{k}: {v}" for k, v in headers.items())


def format_body(content: Union[str, bytes, None], prefix: str) -> str:
    from .conf import settings

    if settings.LOG_OUTGOING_REQUESTS_EMIT_BODY:
        return f"\n{prefix} body:\n{content}"
    return ""


def format_request(req: PreparedRequest) -> str:
    template = textwrap.dedent(
        """
        ---------------- request ----------------
        {req.method} {req.url}
        {reqhdrs} {request_body}
    """
    )
    return template.format(
        req=req,
        reqhdrs=format_headers(req.headers),
        request_body=format_body(req.body, "Request"),
    )


def format_response(resp: Response):
    template = textwrap.dedent(
        """
        ---------------- response ----------------
        {resp.status_code} {resp.reason} {resp.url}
        {reshdrs} {response_body}

    """
    )
    return template.format(
        resp=resp,
        reshdrs=format_headers(resp.headers),
        response_body=format_body(resp.content, "Response"),
    )


def format_error(exception: RequestException) -> str:
    template = textwrap.dedent(
        """
        ---------------- error ----------------
        {msg}

        {tb}
    """
    )
    tb = "\n".join(format_exception(exception))
    output = template.format(msg=str(exception), tb=tb)
    if (request := exception.request) is None:
        return output

    # we have request information, let's include it
    formatted_request = format_request(request)
    return f"{formatted_request}\n{output}"


class HttpFormatter(logging.Formatter):
    """
    Display request and response (meta) details of python requests library log records.

    Depending on the configuration, either only the metadata or metadata + body of
    requests and matching responses is emitted.

    Metadata:

        * HTTP method
        * Request URL
        * Request headers (masking auth details)
        * Response status
        * Response reason
        * Response headers
    """

    def _formatMessageWithResponse(self, record: RequestLogRecord) -> str:
        assert record.req is not None
        assert record.res is not None
        return "{request}\n{response}".format(
            request=format_request(record.req),
            response=format_response(record.res),
        )

    def formatMessage(self, record):
        result = super().formatMessage(record)
        if not is_request_log_record(record):
            return result

        # if there is a response, apply the happy-flow formatting
        if getattr(record, "res", None) is not None:
            record = cast(RequestLogRecord, record)
            output = self._formatMessageWithResponse(record)
            return f"{result}{output}"

        # if there is exception information, extract the details from that
        if record.exc_info and isinstance(
            (exception := record.exc_info[1]), RequestException
        ):
            output = format_error(exception)
            return f"{result}{output}"

        # any other log record - use the default formatter
        return result



import logging
from types import TracebackType
from typing import Optional, Tuple, Type, Union

from requests import RequestException
from requests.models import PreparedRequest, Response


class RequestLogRecord(logging.LogRecord):
    req: Optional[PreparedRequest]
    res: Response


class ErrorRequestLogRecord(logging.LogRecord):
    exc_info: Tuple[Type[BaseException], RequestException, Optional[TracebackType]]


AnyLogRecord = Union[logging.LogRecord, RequestLogRecord, ErrorRequestLogRecord]


def is_request_log_record(record: AnyLogRecord) -> bool:
    return getattr(record, "_is_log_outgoing_requests", False)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class LogOutgoingRequestsConfig(AppConfig):
    name = "log_outgoing_requests"
    verbose_name = _("Outgoing request logs")
    default_auto_field = "django.db.models.BigAutoField"

    def ready(self):
        from .log_requests import install_outgoing_requests_logging

        install_outgoing_requests_logging()



"""Tests for the admin interface"""

from django.urls import reverse
from django.utils import timezone

import pytest
import requests
from pyquery import PyQuery

from log_outgoing_requests.models import OutgoingRequestsLog, OutgoingRequestsLogConfig


#
# test display
#
@pytest.mark.django_db
def test_decoded_content_display(admin_client):
    """Assert that decoded request/response bodies are properly displayed"""

    log = OutgoingRequestsLog.objects.create(
        id=1,
        req_body=b"I'm a lumberjack and I'm okay.",
        res_body=b"I sleep all night and work all day.",
        timestamp=timezone.now(),
    )
    url = reverse(
        "admin:log_outgoing_requests_outgoingrequestslog_change", args=(log.pk,)
    )

    response = admin_client.get(url)
    assert response.status_code == 200

    html = response.content.decode("utf-8")
    doc = PyQuery(html)
    request_body = doc.find(".field-request_body .readonly").text()
    response_body = doc.find(".field-response_body .readonly").text()

    assert request_body == "I'm a lumberjack and I'm okay."
    assert response_body == "I sleep all night and work all day."


#
# test override of settings
#
@pytest.mark.django_db
def test_admin_override_db_save(requests_mock, request_mock_kwargs):
    """Assert that saving logs can be disabled in admin"""

    config = OutgoingRequestsLogConfig.get_solo()
    config.save_to_db = "no"
    config.save()

    requests_mock.get(**request_mock_kwargs)
    requests.get(
        request_mock_kwargs["url"], headers=request_mock_kwargs["request_headers"]
    )

    request_log = OutgoingRequestsLog.objects.last()

    assert request_log is None


@pytest.mark.django_db
def test_admin_override_save_body(requests_mock, request_mock_kwargs):
    """Assert that saving body can be disabled in admin"""

    config = OutgoingRequestsLogConfig.get_solo()
    config.save_body = "no"
    config.save()

    requests_mock.get(**request_mock_kwargs)
    requests.get(
        request_mock_kwargs["url"], headers=request_mock_kwargs["request_headers"]
    )

    request_log = OutgoingRequestsLog.objects.last()

    assert request_log.res_body == b""


@pytest.mark.django_db
def test_admin_override_max_content_length(requests_mock, request_mock_kwargs):
    """Assert that `max_content_length` can be overriden in admin"""

    config = OutgoingRequestsLogConfig.get_solo()
    config.max_content_length = "10"
    config.save()

    requests_mock.get(**request_mock_kwargs)
    requests.get(
        request_mock_kwargs["url"], headers=request_mock_kwargs["request_headers"]
    )

    request_log = OutgoingRequestsLog.objects.last()

    assert request_log.res_body == b""


@pytest.mark.django_db
def test_list_url_is_trucated_over_200_chars(admin_client):
    OutgoingRequestsLog.objects.create(
        id=1,
        url="https://example.com/a1b2c3d4e5f6g7h8i9j0k1l2m3n4o5p6q7r8s9t1u2v3w4x5y6z1a2b3c4d5e6f7g8h9i0j1k2l3m4n5o6p7q8r9s1t2u3v4w5x6y7z/some-path-3894ndjidjd93djjd3eu9jjddu9eu93j3e39ei9idjd3ddksdj9393/some-path/some-path-as-well/skdlkdlskdksdkd9828393jdd",
        timestamp=timezone.now(),
    )
    url = reverse("admin:log_outgoing_requests_outgoingrequestslog_changelist")

    response = admin_client.get(url)
    html = response.content.decode("utf-8")
    doc = PyQuery(html)
    truncated_url = doc.find(".field-truncated_url").text()

    assert (
        truncated_url
        == "/a1b2c3d4e5f6g7h8i9j0k1l2m3n4o5p6q7r8s9t1u2v3w4x5y6z1a2b3c4d5e6f7g8h9i0j1k2l3m4n5o6p7q8r9s1t2u3v4w \u2026 d93djjd3eu9jjddu9eu93j3e39ei9idjd3ddksdj9393/some-path/some-path-as-well/skdlkdlskdksdkd9828393jdd"
    )


@pytest.mark.django_db
def test_list_url_is_not_trucated_under_200_chars(admin_client):
    OutgoingRequestsLog.objects.create(
        id=1,
        url="https://example.com/a1b2c3d4e/some-path",
        timestamp=timezone.now(),
    )
    url = reverse("admin:log_outgoing_requests_outgoingrequestslog_changelist")

    response = admin_client.get(url)
    html = response.content.decode("utf-8")
    doc = PyQuery(html)
    truncated_url = doc.find(".field-truncated_url").text()

    assert truncated_url == "/a1b2c3d4e/some-path"



# Generated by Django 4.1 on 2023-06-09 09:35

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("log_outgoing_requests", "0003_alter_outgoingrequestslog_options_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="hostname",
            field=models.CharField(
                help_text="The netloc/hostname part of the url.",
                max_length=255,
                verbose_name="Hostname",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="req_body_encoding",
            field=models.CharField(
                blank=True,
                help_text="The encoding is either extracted from the Content-Type header or, if absent, taken from the default encoding configured for the content type. If the decoded request content is not displaying correctly, you may try changing the encoding value here.",
                max_length=24,
                verbose_name="Request encoding",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="req_content_type",
            field=models.CharField(
                help_text="The content type of the request.",
                max_length=50,
                verbose_name="Request content type",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="req_headers",
            field=models.TextField(
                help_text="The request headers.", verbose_name="Request headers"
            ),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="res_body_encoding",
            field=models.CharField(
                blank=True,
                help_text="The encoding is either extracted from the Content-Type header or, if absent, taken from the default encoding configured for the content type. If the decoded response content is not displaying correctly, you may try changing the encoding value here.",
                max_length=24,
                verbose_name="Response encoding",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="res_content_type",
            field=models.CharField(
                help_text="The content type of the response.",
                max_length=50,
                verbose_name="Response content type",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="res_headers",
            field=models.TextField(
                help_text="The response headers.", verbose_name="Response headers"
            ),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="trace",
            field=models.TextField(
                help_text="Text providing information in case of request failure.",
                verbose_name="Trace",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="url",
            field=models.URLField(
                help_text="The url of the outgoing request.",
                max_length=1000,
                verbose_name="URL",
            ),
        ),
    ]



from urllib.parse import urlparse

from django import forms
from django.contrib import admin
from django.utils.translation import gettext as _

from solo.admin import SingletonModelAdmin

from .conf import settings
from .models import OutgoingRequestsLog, OutgoingRequestsLogConfig


@admin.register(OutgoingRequestsLog)
class OutgoingRequestsLogAdmin(admin.ModelAdmin):
    list_display = (
        "hostname",
        "truncated_url",
        "params",
        "status_code",
        "method",
        "response_ms",
        "timestamp",
    )
    list_filter = ("method", "timestamp", "status_code", "hostname")
    search_fields = ("url", "params", "hostname")
    date_hierarchy = "timestamp"
    show_full_result_count = False

    fieldsets = (
        (
            _("Request"),
            {
                "fields": (
                    "method",
                    "url",
                    "timestamp",
                    "query_params",
                    "params",
                    "req_headers",
                    "req_content_type",
                    "req_body_encoding",
                    "request_body",
                )
            },
        ),
        (
            _("Response"),
            {
                "fields": (
                    "status_code",
                    "response_ms",
                    "res_headers",
                    "res_content_type",
                    "res_body_encoding",
                    "response_body",
                )
            },
        ),
        (_("Extra"), {"fields": ("trace",)}),
    )
    readonly_fields = (
        "url",
        "timestamp",
        "method",
        "query_params",
        "params",
        "req_headers",
        "req_content_type",
        "request_body",
        "status_code",
        "response_ms",
        "res_headers",
        "res_content_type",
        "response_body",
        "trace",
    )

    class Media:
        css = {
            "all": ("log_outgoing_requests/css/admin.css",),
        }

    def has_add_permission(self, request):
        return False

    @admin.display(description=_("Query parameters"))
    def query_params(self, obj):
        return obj.query_params

    @admin.display(description=_("Request body"))
    def request_body(self, obj) -> str:
        return obj.request_body_decoded or "-"

    @admin.display(description=_("Response body"))
    def response_body(self, obj) -> str:
        return obj.response_body_decoded or "-"

    def truncated_url(self, obj):
        parsed_url = urlparse(obj.url)
        path = parsed_url.path
        max_length = 200
        path_length = len(path)

        if path_length <= max_length:
            return path

        half_length = (max_length - 3) // 2
        left_half = path[:half_length]
        right_half = path[-half_length:]
        return left_half + " \u2026 " + right_half


class ConfigAdminForm(forms.ModelForm):
    class Meta:
        model = OutgoingRequestsLogConfig
        fields = "__all__"
        help_texts = {
            "save_to_db": _(
                "Whether request logs should be saved to the database (default: {default})."
            ).format(default=settings.LOG_OUTGOING_REQUESTS_DB_SAVE),
            "save_body": _(
                "Whether the body of the request and response should be logged (default: "
                "{default})."
            ).format(default=settings.LOG_OUTGOING_REQUESTS_DB_SAVE_BODY),
        }


@admin.register(OutgoingRequestsLogConfig)
class OutgoingRequestsLogConfigAdmin(SingletonModelAdmin):
    form = ConfigAdminForm



from django.conf import settings

from appconf import AppConf

from .datastructures import ContentType

__all__ = ["settings"]


class LogOutgoingRequestsConf(AppConf):
    """
    Settings for django-log-outgoing-requests.

    To override a setting, prefix it with ``LOG_OUTGOING_REQUESTS_`` in your own
    settings file.
    """

    DB_SAVE = False
    """
    Whether request logs should be saved to the database.

    This can be overridden at runtime via configuration in the admin.
    """
    DB_SAVE_BODY = False
    """
    Whether request/response bodies should be saved to the database.

    This can be overridden at runtime via configuration in the admin.
    """
    CONTENT_TYPES = [
        ContentType(pattern="application/json", default_encoding="utf-8"),
        ContentType(pattern="application/soap+xml", default_encoding="utf-8"),
        ContentType(pattern="application/xml", default_encoding="utf-8"),
        ContentType(pattern="text/xml", default_encoding="iso-8859-1"),
        ContentType(pattern="text/*", default_encoding="utf-8"),
    ]
    """
    Allowlist of content types for which the body may be saved to the database.

    Use :class:`log_outgoing_requests.datastructures.ContentType` to configure this
    setting.
    """
    EMIT_BODY = False
    """
    Whether request/response body may be emitted in the logs.
    """
    MAX_CONTENT_LENGTH = 524_288  # 0.5 MB
    """
    The maximum size of request/response bodies for saving to the database, in bytes.

    If the body is larger than this treshold, the log record will still be saved to the
    database, but the body will be missing.
    """

    MAX_AGE = 1
    """
    The maximum age (in days) of request logs, after which they are deleted (via a Celery
    task, Django management command, or the like).
    """

    RESET_DB_SAVE_AFTER = 60
    """
    If the config has been updated, reset the database logging after the specified
    number of minutes.

    To protect against unintended logging of potentially sensitive data after debugging,
    this helps in resetting the "save to DB" configuration option. It resets back to
    "use the default from settings".

    If the value is falsy (including zero), then no reset takes place at all.

    .. note:: this requires Celery to be installed, an optional dependency.
    """



# Generated by Django 4.1 on 2023-10-06 12:08

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("log_outgoing_requests", "0004_alter_outgoingrequestslog_hostname_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="url",
            field=models.TextField(
                help_text="The url of the outgoing request.",
                verbose_name="URL",
            ),
        ),
    ]
